/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.scsm.mms;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.openiec61850.common.ServiceError;
import org.openiec61850.common.model.Array;
import org.openiec61850.common.model.BasicDataAttribute;
import org.openiec61850.common.model.ConstructedDataAttribute;
import org.openiec61850.common.model.DataObject;
import org.openiec61850.common.model.DataSet;
import org.openiec61850.common.model.FunctionalConstraint;
import org.openiec61850.common.model.ModelNode;
import org.openiec61850.common.model.datypes.basic.BOOLEAN;
import org.openiec61850.common.model.datypes.basic.CodedEnum;
import org.openiec61850.common.model.datypes.basic.Enumerated;
import org.openiec61850.common.model.datypes.basic.FLOAT32;
import org.openiec61850.common.model.datypes.basic.FLOAT64;
import org.openiec61850.common.model.datypes.basic.INT16;
import org.openiec61850.common.model.datypes.basic.INT16U;
import org.openiec61850.common.model.datypes.basic.INT32;
import org.openiec61850.common.model.datypes.basic.INT32U;
import org.openiec61850.common.model.datypes.basic.INT64;
import org.openiec61850.common.model.datypes.basic.INT8;
import org.openiec61850.common.model.datypes.basic.INT8U;
import org.openiec61850.common.model.datypes.basic.OCTET_STRING;
import org.openiec61850.common.model.datypes.basic.UNICODE_STRING;
import org.openiec61850.common.model.datypes.basic.VISIBLE_STRING;
import org.openiec61850.common.model.datypes.commonacsi.Check;
import org.openiec61850.common.model.datypes.commonacsi.Quality;
import org.openiec61850.common.model.datypes.commonacsi.Quality.Source;
import org.openiec61850.common.model.datypes.commonacsi.Quality.Validity;
import org.openiec61850.common.model.datypes.commonacsi.Timestamp;
import org.openiec61850.common.model.report.OptFields;
import org.openiec61850.common.model.report.ReportControlBlock;
import org.openiec61850.common.model.report.TriggerConditions;
import org.openiec61850.jmms.mms.asn1.Data;
import org.openiec61850.jmms.mms.asn1.DataSequence;
import org.openmuc.jasn1.ber.types.BerBitString;
import org.openmuc.jasn1.ber.types.BerBoolean;
import org.openmuc.jasn1.ber.types.BerInteger;
import org.openmuc.jasn1.ber.types.BerOctetString;
import org.openmuc.jasn1.ber.types.string.BerUTF8String;
import org.openmuc.jasn1.ber.types.string.BerVisibleString;

public class MmsAsn1ComplexTypeConverter {

	/**
	 * Numbers can be plain numbers like Integer, but some Enumerated data types
	 * are also implemented as numbers. In java they can be implemented using
	 * {@link Enum}s. In this case we take the ordinal value of the Enum value
	 * as the integer value.
	 * 
	 * This does not work in all cases. It is only possible, if the Enumerated
	 * values start with 0 and have no holes.
	 */
	private static Long getNumberValue(BasicDataAttribute da) throws ServiceError {
		Object rawValue = da.getValue();
		Long value;
		if (rawValue == null) {
			value = 0L;
		}
		else if (rawValue instanceof Enum<?>) {
			value = Long.valueOf(((Enum<?>) rawValue).ordinal());
		}
		else if (rawValue instanceof Number) {
			value = Long.valueOf(((Number) rawValue).longValue());
		}
		else {
			throw new ServiceError(ServiceError.TYPE_CONFLICT, "value of " + da.getReference() + " should be "
					+ da.getBasicType() + " but is " + rawValue.getClass());
		}
		return value;
	}

	public static Data convertModelNodeToMmsData(ModelNode node) throws ServiceError {

		if (node instanceof BasicDataAttribute) {
			return convertBasicNodeToMmsData((BasicDataAttribute) node);
		}
		else if (node instanceof Array) {
			return convertArrayNodeToMmsData((Array) node);
		}
		else if (node instanceof DataObject || node instanceof ConstructedDataAttribute
				|| node instanceof ReportControlBlock) {
			return convertStructNodeToMmsData(node);
		}
		else if (node instanceof TriggerConditions || node instanceof OptFields || node instanceof DataSet) {
			return convertNodeToMmsData(node);
		}
		else {
			throw new IllegalArgumentException("invalid ModelNode for: " + node.getNodeName() + " passed.");
		}
	}

	private static Data convertStructNodeToMmsData(ModelNode modelNode) throws ServiceError {

		Collection<ModelNode> subNodes = modelNode.getChildren();
		// Collections.sort(subNodes,
		// MMSNodeSorter.getDataContainerComparator());

		// Data data = new Data();
		LinkedList<Data> seq = new LinkedList<Data>();
		for (ModelNode item : subNodes) {
			Data child = convertModelNodeToMmsData(item);
			if (child == null) {
				continue;
			}
			seq.add(child);
		}
		if (seq.size() == 0) {
			return null;
		}
		// data.selectStructure(new DataSequence(seq));
		// return data;

		DataSequence dataSeq = new DataSequence(seq);
		return new Data(null, dataSeq, null, null, null, null, null, null, null, null, null, null, null, null, null);
	}

	private static Data convertArrayNodeToMmsData(Array array) throws ServiceError {

		ArrayList<Data> mmsArray = new ArrayList<Data>();
		for (ModelNode arrayItem : array) {
			Data mmsArrayItem = convertModelNodeToMmsData(arrayItem);
			if (mmsArrayItem != null) {
				mmsArray.add(mmsArrayItem);
			}
		}

		DataSequence dataSeq = new DataSequence(mmsArray);
		return new Data(dataSeq, null, null, null, null, null, null, null, null, null, null, null, null, null, null);

	}

	private static Data convertNodeToMmsData(ModelNode node) throws ServiceError {
		// Data data = new Data();

		if (node instanceof TriggerConditions) {
			return new Data(null, null, null, new BerBitString(((TriggerConditions) node).getValue(), 6), null, null,
					null, null, null, null, null, null, null, null, null);
		}

		else if (node instanceof OptFields) {
			// data.selectBit_string(new
			// BitString(MmsAsn1PrimitiveTypeConverter.toByteArray((((OptFields)node).getValue()))));
			// data.selectBit_string(new BitString(((OptFields)
			// node).getValue(),
			// 6));
			return new Data(null, null, null, new BerBitString(((OptFields) node).getValue(), 10), null, null, null,
					null, null, null, null, null, null, null, null);
		}

		// TODO check if data sets may be handled differently
		else {// data set
				// data.selectVisible_string(node.getReference().toString());
			try {
				return new Data(null, null, null, null, null, null, null, null, new BerVisibleString(node
						.getReference().toString()), null, null, null, null, null, null);

			} catch (UnsupportedEncodingException e) {
				throw new IllegalAccessError();
			}
		}

	}

	private static Data convertBasicNodeToMmsData(BasicDataAttribute da) throws ServiceError {

		if (da.getValue() == null) {
			return null;
		}
		// Data data = null;
		Object daValue = da.getValue();

		switch (da.getBasicType()) {
		case BOOLEAN:
			// data.selectBoolean_((Boolean) daValue);
			return new Data(null, null, new BerBoolean((Boolean) daValue), null, null, null, null, null, null, null,
					null, null, null, null, null);

		case INT8:
		case INT16:
		case INT32:
		case INT64:
			return new Data(null, null, null, null, new BerInteger(getNumberValue(da)), null, null, null, null, null,
					null, null, null, null, null);
		case INT8U:
		case INT16U:
		case INT32U:
			return new Data(null, null, null, null, null, new BerInteger(getNumberValue(da)), null, null, null, null,
					null, null, null, null, null);
		case FLOAT32:
			return new Data(null, null, null, null, null, null, new BerOctetString(
					MmsAsn1PrimitiveTypeConverter.toByteArray(((Float) daValue))), null, null, null, null, null, null,
					null, null);

		case FLOAT64:
			return new Data(null, null, null, null, null, null, new BerOctetString(
					MmsAsn1PrimitiveTypeConverter.toByteArray(((Double) daValue))), null, null, null, null, null, null,
					null, null);
		case CODEDENUM:
			// data.selectBit_string(new BitString(((byte[]) daValue)));
			// return data;
			// TODO always 2 bits??

			byte[] value = (byte[]) daValue;

			return new Data(null, null, null, new BerBitString(value, value.length * 8), null, null, null, null, null,
					null, null, null, null, null, null);

		case ENUMERATED:
			return new Data(null, null, null, null, new BerInteger(getNumberValue(da)), null, null, null, null, null,
					null, null, null, null, null);

		case OCTET_STRING:

			if (((OCTET_STRING) da).getMaxLength() == 6) {
				if (daValue == null) {
					daValue = new byte[0];
				}
				if (da.getNodeName().equals("TimeOfEntry")) {
					// data.selectBinary_time(new TimeOfDay((byte[]) daValue));
					return new Data(null, null, null, null, null, null, null, null, null, null, new BerOctetString(
							(byte[]) daValue), null, null, null, null);
				}
				else {// other values with OCTET_STRING_6??
						// data.selectOctet_string((byte[]) daValue);
					return new Data(null, null, null, null, null, null, null, new BerOctetString((byte[]) daValue),
							null, null, null, null, null, null, null);
				}
			}

			if (daValue == null) {
				daValue = new byte[0];
			}
			return new Data(null, null, null, null, null, null, null, new BerOctetString((byte[]) daValue), null, null,
					null, null, null, null, null);
		case VISIBLE_STRING: // fall through
			if (daValue == null) {
				daValue = "";
			}
			try {
				return new Data(null, null, null, null, null, null, null, null, new BerVisibleString((String) daValue),
						null, null, null, null, null, null);
			} catch (UnsupportedEncodingException e) {
				throw new IllegalAccessError();
			}
		case UNICODE_STRING:
			if (daValue == null) {
				daValue = "";
			}
			return new Data(null, null, null, null, null, null, null, null, null, null, null, null, null,
					new BerUTF8String(((String) daValue).getBytes()), null);
		case TIMESTAMP:

			return new Data(null, null, null, null, null, null, null, null, null, null, null, null, null, null,
					new BerOctetString(MmsAsn1PrimitiveTypeConverter.toByteArray((Timestamp) da)));
		case QUALITY:

			byte[] qvalue = MmsAsn1PrimitiveTypeConverter.toByteArray((Quality) da);

			return new Data(null, null, null, new BerBitString(qvalue, qvalue.length * 8), null, null, null, null,
					null, null, null, null, null, null, null);
		case CHECK:
			byte[] checkValue = new byte[] { 0 };
			if (((Check) da).synchrocheck) {
				checkValue[0] = (byte) 0x80;
			}
			if (((Check) da).interlock_check) {
				checkValue[0] |= (byte) 0x40;
			}
			return new Data(null, null, null, new BerBitString(checkValue, 2), null, null, null, null, null, null,
					null, null, null, null, null);
		default:
			throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);

		}
	}

	/**
	 * Converts MMS raw data to a ModelNode. At the moment it doesn't reuse but
	 * recreates all model nodes. When fc != null, the resulting ModelNode only
	 * contains children matching the FunctionalConstraint.
	 * 
	 * Does not write the new values into the model only the copy. ModelNode dd
	 * is already a copy!
	 * 
	 */
	public static void convertMmsDataToDataContainer(ModelNode dd, FunctionalConstraint fc, Data data, String[] mmsFCs)
			throws ServiceError {

		if (data.boolean_ != null) {
			if (!(dd instanceof BOOLEAN)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}
			((BOOLEAN) dd).setValue(data.boolean_.val);
		}
		else if (data.bit_string != null) {
			if (!(dd instanceof CodedEnum) && !(dd instanceof Quality) && !(dd instanceof TriggerConditions)
					&& !(dd instanceof OptFields) && !(dd instanceof Check)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			byte[] rawData = data.bit_string.bitString;
			if (dd instanceof BasicDataAttribute) {

				switch (((BasicDataAttribute) dd).getBasicType()) {
				case TIMESTAMP:
					toTimestamp((Timestamp) dd, rawData);
					break;
				case QUALITY:
					toQuality((Quality) dd, rawData);
					break;
				case CHECK:
					break;
				// case DBPOS:
				// return toDbpos((Dbpos)dd, rawData);
				// case TCMD:
				case CODEDENUM:
					toCodedEnum((CodedEnum) dd, rawData);
					break;
				default:
					throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE, "invalid or unknown BasicType: "
							+ ((BasicDataAttribute) dd).getBasicType());
				}

			}
			else if (dd instanceof OptFields) {
				((OptFields) dd).setValue(rawData);
			}
			else if (dd instanceof TriggerConditions) {
				((TriggerConditions) dd).setValue(rawData);
			}
		}
		else if (data.integer != null) {

			if (dd instanceof INT8) {
				((INT8) dd).setValue((byte) data.integer.val);
			}

			else if (dd instanceof INT16) {
				((INT16) dd).setValue((short) data.integer.val);
			}

			else if (dd instanceof INT32) {
				((INT32) dd).setValue((int) data.integer.val);
			}

			else if (dd instanceof INT64) {
				((INT64) dd).setValue(data.integer.val);
			}

			else if (dd instanceof Enumerated) {
				((Enumerated) dd).setValue((int) data.integer.val);
			}

			else {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}
		}

		else if (data.unsigned != null) {
			if (dd instanceof INT8U) {
				((INT8U) dd).setValue((byte) data.unsigned.val);
			}

			else if (dd instanceof INT16U) {
				((INT16U) dd).setValue((short) data.unsigned.val);
			}

			else if (dd instanceof INT32U) {
				((INT32U) dd).setValue((int) data.unsigned.val);
			}

			else {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}
		}

		else if (data.floating_point != null) {
			if (!(dd instanceof FLOAT32) && !(dd instanceof FLOAT64)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			byte[] rawData = data.floating_point.octetString;

			if (dd instanceof FLOAT32) {

				if (rawData == null || rawData.length != 5) {
					throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
				}
				byte[] b2 = new byte[4];
				System.arraycopy(rawData, 1, b2, 0, 4);

				((FLOAT32) dd).setValue(Float.intBitsToFloat(((0xff & b2[0]) << 24 | (0xff & b2[1]) << 16
						| (0xff & b2[2]) << 8 | (0xff & b2[3]) << 0)));

			}

			else if (dd instanceof FLOAT64) {

				if (rawData == null || rawData.length != 9) {
					throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
				}
				byte[] b2 = new byte[8];
				System.arraycopy(rawData, 1, b2, 0, 8);

				((FLOAT64) dd).setValue(Double.longBitsToDouble((0xff & b2[0]) << 56 | (0xff & b2[1]) << 48
						| (0xff & b2[2]) << 40 | (0xff & b2[3]) << 32 | (0xff & b2[4]) << 24 | (0xff & b2[5]) << 16
						| (0xff & b2[6]) << 8 | (0xff & b2[7]) << 0));

			}

		}

		else if (data.octet_string != null) {

			if (!(dd instanceof OCTET_STRING)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			((OCTET_STRING) dd).setValue(data.octet_string.octetString);

		}
		else if (data.visible_string != null) {

			if (!(dd instanceof VISIBLE_STRING)) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			((VISIBLE_STRING) dd).setValue(data.visible_string.toString());

		}
		else if (data.generalized_time != null) {
			// TODO type not implemented
		}
		else if (data.binary_time != null) {
			((OCTET_STRING) dd).setValue(data.binary_time.octetString);

		}
		else if (data.bcd != null) {
			// TODO type not implemented
		}
		else if (data.booleanArray != null) {
			// TODO type not implemented
		}
		else if (data.array != null && dd instanceof Array) {

			List<Data> items = data.array.seqOf;
			List<ModelNode> childNodes = (ArrayList<ModelNode>) ((Array) dd).getChildren();

			if (childNodes.size() != items.size()) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INAPPROPRIATE);
			}

			int i = 0;
			for (Data item : items) {
				convertMmsDataToDataContainer(childNodes.get(i), fc, item, mmsFCs);
				i++;
			}

		}
		else if (data.mMSString != null) {
			((UNICODE_STRING) dd).setValue(data.mMSString.toString());

		}
		else if (data.utc_time != null) {
			toTimestamp((Timestamp) dd, data.utc_time.octetString);
		}
		else if (data.structure != null) {

			List<ModelNode> result = new ArrayList<ModelNode>();
			DataSequence dataSeq = data.structure;
			Iterator<Data> resIter = dataSeq.seqOf.iterator();

			if (dd.getChildren() == null) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"mmsDataToDataContainer: Structure must contain at least one element.");
			}

			List<ModelNode> sortedChildren = new LinkedList<ModelNode>();
			FunctionalConstraint fcTemp = null;
			for (String inOrderFc : mmsFCs) {
				for (ModelNode child : dd.getChildren()) {
					if (child instanceof DataSet) {
						continue;
					}

					fcTemp = FunctionalConstraint.fromString(inOrderFc);
					if (fcTemp == null) {
						continue; // for cases of RP, BR, LG etc
					}

					if (fcTemp.equals(child.getFunctionalConstraint())) {
						if ((fc == null || fc.equals(child.getFunctionalConstraint()))) {
							sortedChildren.add(child);
						}
					}
				}
			}

			for (ModelNode childDef : sortedChildren) {
				if (!resIter.hasNext()) {
					throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
							"mmsDataToDataContainer: Wrong number of results received (A).");
				}

				Data subData = resIter.next();
				// ModelNode da = convertMmsDataToDataContainer(childDef, fc,
				// subData, mmsFCs);
				convertMmsDataToDataContainer(childDef, fc, subData, mmsFCs);
				result.add(childDef);
			}

			if (resIter.hasNext()) {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT,
						"mmsDataToDataContainer: Wrong number of results received (B).");
			}

			if (dd instanceof DataObject) {
				dd = new DataObject(dd.getReference(), result, "");
				// TODO ok to leave cdcName null?
			}

			else if (dd instanceof ConstructedDataAttribute) {
				dd = new ConstructedDataAttribute(dd.getReference(), dd.getFunctionalConstraint(), result);

			}

			else {
				throw new ServiceError(ServiceError.PARAMETER_VALUE_INCONSISTENT, "Unknown or invalid ObjectType for: "
						+ dd.getNodeName());
			}

		}
		else {
			throw new ServiceError(ServiceError.CLASS_NOT_SUPPORTED, "Response contained unknown type");
		}
		// TODO correct error code?
	}

	private static void toCodedEnum(CodedEnum dd, byte[] rawData) throws ServiceError {
		// CodedEnum codedEnum = (CodedEnum)dd;
		((CodedEnum) dd).setValue(rawData); // need to do more to the rawData?
		// return codedEnum;
	}

	/**
	 * see {@link #toByteArray(Quality)}
	 */
	private static void toQuality(Quality q, byte[] rawData) {
		byte validity = (byte) (rawData[0] & 0x03);
		switch (validity) {
		case 0:
			q.validity = Validity.GOOD;
			break;
		case 1:
			q.validity = Validity.INVALID;
			break;
		case 2:
			q.validity = Validity.RESERVED;
			break;
		case 3:
			q.validity = Validity.QUESTIONABLE;
			break;
		}
		q.overflow = (rawData[0] & 0x04) != 0;
		q.outOfRange = (rawData[0] & 0x08) != 0;
		q.badReference = (rawData[0] & 0x10) != 0;
		q.oscillatory = (rawData[0] & 0x20) != 0;
		q.failure = (rawData[0] & 0x40) != 0;
		q.oldData = (rawData[0] & 0x80) != 0;
		q.inconsistent = (rawData[1] & 0x01) != 0;
		q.inaccurate = (rawData[1] & 0x02) != 0;
		q.source = (rawData[1] & 0x04) != 0 ? Source.SUBSTITUTED : Source.PROCESS;
		q.test = (rawData[1] & 0x08) != 0;
		q.operatorBlocked = (rawData[1] & 0x10) != 0;
	}

	public static void toTimestamp(Timestamp t, byte[] b) {

		t.secondsSinceEpoch = (0xffL & b[0]) << 24 | (0xffL & b[1]) << 16 | (0xffL & b[2]) << 8 | (0xffL & b[3]);
		t.fractionOfSecond = (0xff & b[4]) << 16 | (0xff & b[5]) << 8 | (0xff & b[6]);
		t.timeQuality.leapSecondsknown = (b[7] & 0x80) != 0;
		t.timeQuality.clockFailure = (b[7] & 0x40) != 0;
		t.timeQuality.clockNotSynchronized = (b[7] & 0x20) != 0;
		t.timeQuality.timeAccuracy = (int) (b[7] & 0x1f);

	}

}
