/*
 * Copyright Fraunhofer ISE, energy & meteo Systems GmbH, and other contributors 2011
 *
 * This file is part of openIEC61850.
 * For more information visit http://www.openmuc.org 
 *
 * openIEC61850 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * openIEC61850 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with openIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openiec61850.scsm.mms;

import java.nio.ByteBuffer;

import org.openiec61850.common.model.datypes.commonacsi.Quality;
import org.openiec61850.common.model.datypes.commonacsi.Timestamp;


/**
 * @author tweidelt
 * @author Stefan Feuerhahn
 */
public class MmsAsn1PrimitiveTypeConverter {
	public static final byte MMS_FLOAT32_EXPONENT_WIDTH = 8;
	public static final byte MMS_FLOAT64_EXPONENT_WIDTH = 11;

	public static byte toByte(byte[] b) {
		return (b == null || b.length == 0) ? 0x0 : b[0];
	}

	public static byte[] toByteArray(boolean value) {
		return new byte[] { (byte) (value ? 0x01 : 0x00) };
	}

	public static byte[] toByteArray(byte value) {
		return new byte[] { value };
	}

	/**
	 * Converts a double value to a 9-byte array. First byte contains the
	 * exponent width, normally 11 bits.
	 */
	public static byte[] toByteArray(double value) {
		return ByteBuffer.allocate(1 + 8).put(MMS_FLOAT64_EXPONENT_WIDTH).putDouble(value).array();
	}

	/**
	 * Converts a float value to a 5-byte array. First byte contains the
	 * exponent width, normally 8 bits.
	 */
	public static byte[] toByteArray(float value) {
		return ByteBuffer.allocate(1 + 4).put(MMS_FLOAT32_EXPONENT_WIDTH).putFloat(value).array();
	}

	public static byte[] toByteArray(long value) {
		return new byte[] { (byte) ((value >> 56) & 0xff), (byte) ((value >> 48) & 0xff),
				(byte) ((value >> 40) & 0xff), (byte) ((value >> 32) & 0xff), (byte) ((value >> 24) & 0xff),
				(byte) ((value >> 16) & 0xff), (byte) ((value >> 8) & 0xff), (byte) ((value >> 0) & 0xff) };
	}

	/**
	 * See 61850-8-1 section
	 * "Mapping of quality common data attribute type specified in IEC 61850-7-3"
	 * in chapter "Common Data Types"
	 */
	public static byte[] toByteArray(Quality value) {
		byte[] result = new byte[2];
		switch (value.validity) {
		case GOOD:
			result[0] = (byte) (result[0] | (byte) 0x00);
			break;
		case INVALID:
			result[0] = (byte) (result[0] | (byte) 0x01);
			break;
		case QUESTIONABLE:
			result[0] = (byte) (result[0] | (byte) 0x02);
			break;
		case RESERVED:
			result[0] = (byte) (result[0] | (byte) 0x03);
			break;
		}
		if (value.overflow) {
			result[0] = (byte) (result[0] | (byte) 0x04);
		}
		if (value.outOfRange) {
			result[0] = (byte) (result[0] | (byte) 0x08);
		}
		if (value.badReference) {
			result[0] = (byte) (result[0] | (byte) 0x10);
		}
		if (value.oscillatory) {
			result[0] = (byte) (result[0] | (byte) 0x20);
		}
		if (value.failure) {
			result[0] = (byte) (result[0] | (byte) 0x40);
		}
		if (value.oldData) {
			result[0] = (byte) (result[0] | (byte) 0x80);
		}
		if (value.inconsistent) {
			result[1] = (byte) (result[1] | (byte) 0x01);
		}
		if (value.inaccurate) {
			result[1] = (byte) (result[1] | (byte) 0x02);
		}
		switch (value.source) {
		case PROCESS:
			result[1] = (byte) (result[1] | (byte) 0x00);
			break;
		case SUBSTITUTED:
			result[1] = (byte) (result[1] | (byte) 0x04);
			break;
		}
		if (value.test) {
			result[1] = (byte) (result[1] | (byte) 0x08);
		}
		if (value.operatorBlocked) {
			result[1] = (byte) (result[1] | (byte) 0x10);
		}
		return result;
	}

	public static byte[] toByteArray(short value) {
		return new byte[] { (byte) ((value >> 8) & 0xff), (byte) ((value >> 0) & 0xff) };
	}

	// public static byte[] toByteArray(Tcmd value) {
	// // FIXME Auto-generated method stub
	// return new byte[1];
	// }

	// TODO needs testing!
	/**
	 * See 61850-8-1 Section "TIMESTAMP" in chapter "Common data types"
	 */
	// public static OCTET_STRING_64 toMmsUtcTime(Timestamp value) {
	// // long quality = value.timeQuality.timeAccuracy & 0x001f;
	// // if (value.timeQuality.leapSecondsknown) {
	// // quality = quality | 0x80;
	// // }
	// // if (value.timeQuality.clockFailure) {
	// // quality = quality | 0x40;
	// // }
	// // if (value.timeQuality.clockNotSynchronized) {
	// // quality = quality | 0x20;
	// // }
	// // long timestamp = (value.secondsSinceEpoch << 32) & 0xffffffff
	// // | (value.fractionOfSecond << 8) | quality;
	//
	// OCTET_STRING_64 mmsUtcTime = new OCTET_STRING_64(value.getReference(),
	// value.getFunctionalConstraint(), "",
	// toByteArray(value), false, false);
	//
	// // UtcTime mmsUtcTime = new UtcTime();
	// // mmsUtcTime.setValue(toByteArray(timestamp));
	// // mmsUtcTime.setValue(toByteArray(value));
	// return mmsUtcTime;
	// }

	public static byte[] toByteArray(Timestamp value) {
		long quality = value.timeQuality.timeAccuracy & 0x001f;
		if (value.timeQuality.leapSecondsknown) {
			quality = quality | 0x80;
		}
		if (value.timeQuality.clockFailure) {
			quality = quality | 0x40;
		}
		if (value.timeQuality.clockNotSynchronized) {
			quality = quality | 0x20;
		}
		long timestamp = (value.secondsSinceEpoch << 32) & 0xffffffff | (value.fractionOfSecond << 8) | quality;

		return toByteArray(timestamp);
	}

}
